@@ -3,12 +3,15 @@ package main
 import (
 	"bytes"
 	"encoding/json"
+	"errors"
 	"fmt"
 	"io"
 	"io/ioutil"
 	"log"
 	"os"
 	"path/filepath"
+	"regexp"
+	"strconv"
 	"strings"
 
 	"github.com/multiprocessio/datastation/runner"
@@ -18,7 +21,7 @@ import (
 )
 
 func resolveContentType(fileExtensionOrContentType string) runner.MimeType {
-	if strings.Contains(fileExtensionOrContentType, "/") {
+	if strings.Contains(fileExtensionOrContentType, string(filepath.Separator)) {
 		return runner.MimeType(fileExtensionOrContentType)
 	}
 
@@ -44,21 +47,40 @@ func evalFileInto(file string, out *os.File) error {
 	return runner.TransformFile(file, runner.ContentTypeInfo{}, out)
 }
 
-func getShape(resultFile, panelId string) *runner.Shape {
-	s, err := runner.ShapeFromFile(resultFile, panelId, 10_000, 100)
-	if err != nil {
-		log.Fatal(err)
-	}
+func getShape(resultFile, panelId string) (*runner.Shape, error) {
+	return runner.ShapeFromFile(resultFile, panelId, runner.DefaultShapeMaxBytesToRead, 100)
+}
 
-	if !runner.ShapeIsObjectArray(*s) {
-		rest := "."
-		if panelId != "" {
-			rest = ": " + panelId + "."
+var tableFileRe = regexp.MustCompile(`({(?P<number>[0-9]+)(((,\s*(?P<numbersinglepath>"(?:[^"\\]|\\.)*\"))?)|(,\s*(?P<numberdoublepath>'(?:[^'\\]|\\.)*\'))?)})|({((((?P<singlepath>"(?:[^"\\]|\\.)*\"))?)|((?P<doublepath>'(?:[^'\\]|\\.)*\'))?)})`)
+
+func rewriteQuery(query string) string {
+	query = strings.ReplaceAll(query, "{}", "DM_getPanel(0)")
+
+	query = tableFileRe.ReplaceAllStringFunc(query, func(m string) string {
+		matchForSubexps := tableFileRe.FindStringSubmatch(m)
+		index := "0"
+		path := ""
+		for i, name := range tableFileRe.SubexpNames() {
+			if matchForSubexps[i] == "" {
+				continue
+			}
+
+			switch name {
+			case "number":
+				index = matchForSubexps[i]
+			case "numberdoublepath", "numbersinglepath", "doublepath", "singlepath":
+				path = matchForSubexps[i]
+			}
 		}
-		log.Fatalf("Input is not an array of objects%s", rest)
-	}
 
-	return s
+		if path != "" {
+			return fmt.Sprintf("DM_getPanel(%s, %s)", index, path)
+		}
+
+		return fmt.Sprintf("DM_getPanel(%s)", index)
+	})
+
+	return query
 }
 
 var Version = "latest"
@@ -89,7 +111,7 @@ Examples:
 
 See the repo for more details: https://github.com/multiprocessio/dsq.`
 
-func main() {
+func _main() error {
 	log.SetFlags(0)
 	runner.Verbose = false
 	var nonFlagArgs []string
@@ -108,7 +130,7 @@ func main() {
 
 		if arg == "-h" || arg == "--help" {
 			log.Println(HELP)
-			return
+			return nil
 		}
 
 		if arg == "-p" || arg == "--pretty" {
@@ -135,12 +157,12 @@ func main() {
 	}
 
 	if len(files) == 0 {
-		log.Fatal("No input files.")
+		return errors.New("No input files.")
 	}
 
 	projectTmp, err := ioutil.TempFile("", "dsq-project")
 	if err != nil {
-		log.Fatal(err)
+		return err
 	}
 	defer os.Remove(projectTmp.Name())
 	project := &runner.ProjectState{
@@ -152,13 +174,20 @@ func main() {
 		},
 	}
 
+	tmpDir, err := os.MkdirTemp("", "dsq")
+	if err != nil {
+		return err
+	}
+	defer os.RemoveAll(tmpDir)
+
+	ec := runner.NewEvalContext(*runner.DefaultSettings, tmpDir)
 	for i := 0; i < len(files); i++ {
 		file := files[i]
 		panelId := uuid.New().String()
-		resultFile := runner.GetPanelResultsFile(project.Id, panelId)
+		resultFile := ec.GetPanelResultsFile(project.Id, panelId)
 		out, err := openTruncate(resultFile)
 		if err != nil {
-			log.Fatal(err)
+			return err
 		}
 
 		readFromStdin := false
@@ -166,32 +195,35 @@ func main() {
 			b, err := ioutil.ReadAll(os.Stdin)
 			if err == nil {
 				if i == len(files)-1 {
-					log.Fatal("Expected file extension or mimetype: e.g. cat x.csv | dsq -s csv, or cat x.csv | dsq -s text/csv")
+					return errors.New("Expected file extension or mimetype: e.g. cat x.csv | dsq -s csv, or cat x.csv | dsq -s text/csv")
 				}
 				mimetype := files[i+1]
-				if !strings.Contains(mimetype, "/") {
+				if !strings.Contains(mimetype, string(filepath.Separator)) {
 					mimetype = string(runner.GetMimeType("x."+mimetype, runner.ContentTypeInfo{}))
 				}
 
 				if mimetype == "" {
-					log.Fatalf("Unknown mimetype or file extension: %s.", mimetype)
+					return fmt.Errorf("Unknown mimetype or file extension: %s.", mimetype)
 				}
 				i += 1
 
 				cti := runner.ContentTypeInfo{Type: string(mimetype)}
 				err := runner.TransformReader(bytes.NewReader(b), "", cti, out)
 				if err != nil {
-					log.Fatal(err)
+					return err
 				}
 
-				s := getShape(resultFile, "")
+				s, err := getShape(resultFile, "")
+				if err != nil {
+					return err
+				}
 
 				project.Pages[0].Panels = append(project.Pages[0].Panels, runner.PanelInfo{
 					ResultMeta: runner.PanelResult{
 						Shape: *s,
 					},
 					Id:   panelId,
-					Name: uuid.New().String(),
+					Name: file,
 				})
 
 				readFromStdin = true
@@ -203,11 +235,14 @@ func main() {
 		if !readFromStdin {
 			err := evalFileInto(file, out)
 			if err != nil {
-				log.Fatal(err)
+				return err
 			}
 		}
 
-		s := getShape(resultFile, file)
+		s, err := getShape(resultFile, file)
+		if err != nil {
+			return err
+		}
 
 		project.Pages[0].Panels = append(project.Pages[0].Panels, runner.PanelInfo{
 			ResultMeta: runner.PanelResult{
@@ -222,32 +257,28 @@ func main() {
 
 	// No query, just dump transformed file directly out
 	if lastNonFlagArg == "" {
-		resultFile := runner.GetPanelResultsFile(project.Id, project.Pages[0].Panels[0].Id)
+		resultFile := ec.GetPanelResultsFile(project.Id, project.Pages[0].Panels[0].Id)
 		fd, err := os.Open(resultFile)
 		if err != nil {
-			log.Fatal(err)
+			return err
 		}
 		defer fd.Close()
 
 		_, err = io.Copy(os.Stdout, fd)
 		if err != nil {
-			log.Fatal(err)
+			return err
 		}
 
-		return
+		return nil
 	}
 
 	connector, err := runner.MakeTmpSQLiteConnector()
 	if err != nil {
-		log.Fatal(err)
+		return err
 	}
 	project.Connectors = append(project.Connectors, *connector)
 
-	query := lastNonFlagArg
-	query = strings.ReplaceAll(query, "{}", "DM_getPanel(0)")
-	for i := 0; i < len(project.Pages[0].Panels); i++ {
-		query = strings.ReplaceAll(query, fmt.Sprintf("{%d}", i), fmt.Sprintf("DM_getPanel(%d)", i))
-	}
+	query := rewriteQuery(lastNonFlagArg)
 	panel := &runner.PanelInfo{
 		Type:    runner.DatabasePanel,
 		Content: query,
@@ -260,30 +291,45 @@ func main() {
 		},
 	}
 
-	err = runner.EvalDatabasePanel(project, 0, panel, nil)
+	err = ec.EvalDatabasePanel(project, 0, panel, nil)
 	if err != nil {
-		log.Fatal(err)
+		if e, ok := err.(*runner.DSError); ok && e.Name == "NotAnArrayOfObjectsError" {
+			rest := "."
+			nth, err := strconv.Atoi(e.TargetPanelId)
+			if err == nil {
+				rest = ": " + files[nth] + "."
+			}
+			return fmt.Errorf("Input is not an array of objects%s", rest)
+		}
+
+		if e, ok := err.(*runner.DSError); ok && e.Name == "UserError" {
+			if e.Message[len(e.Message)-1] != '.' {
+				e.Message += "."
+			}
+			return errors.New(e.Message)
+		}
+		return err
 	}
 
-	resultFile := runner.GetPanelResultsFile(project.Id, panel.Id)
+	resultFile := ec.GetPanelResultsFile(project.Id, panel.Id)
 	fd, err := os.Open(resultFile)
 	if err != nil {
-		log.Fatalf("Could not open results file: %s", err)
+		return fmt.Errorf("Could not open results file: %s", err)
 	}
 
 	if !pretty {
 		// Dump the result to stdout
 		_, err = io.Copy(os.Stdout, fd)
 		if err != nil {
-			log.Fatal(err)
+			return err
 		}
 		fmt.Println()
-		return
+		return nil
 	}
 
-	s, err := runner.ShapeFromFile(resultFile, "results", 10_000, 100)
+	s, err := runner.ShapeFromFile(resultFile, "results", runner.DefaultShapeMaxBytesToRead, 100)
 	if err != nil {
-		log.Fatal(err)
+		return err
 	}
 
 	var columns []string
@@ -299,7 +345,7 @@ func main() {
 	var rows []map[string]interface{}
 	err = dec.Decode(&rows)
 	if err != nil {
-		log.Fatal(err)
+		return err
 	}
 
 	for _, objRow := range rows {
@@ -323,4 +369,12 @@ func main() {
 	}
 
 	table.Render()
+	return nil
+}
+
+func main() {
+	err := _main()
+	if err != nil {
+		log.Fatal(err)
+	}
 }
